DAT objectName         byte "Eight7Segments140905e", 0
CON
{{
  ******* Public Notes *******

  Author: Duane Degn

  An attempt to drive eight 7-segment displays with
  two '595 shift registers.
  For more information see the following thread in the
  Parallax forums.
  http://forums.parallax.com/showthread.php/157074
 
}}

{
  ******* Private Notes *******

  23a Start adapting program to display numbers
  on a 7-segment display
  26a Update program based on schematic posted to the forum.
  27b Add delay and "COMMON_ANODE" constant.
  27c Correct some of the comments and upload to forum.
  27d Change to correspond to Arduino code from Robot Shop.
  Change name from "Shift7Segment140905a" to "Eight7Segments140905a"
  140905a Swapped the order of the digits and removed the delay between
  digits. Works as expected.
  05b Begin making the program a child object.
  05b Still works using temporary "MainLoop" method.
  05b Pad feature doesn't work yet.
  05e Remove stand alone features.
  
}
CON                     

  DIGITS = 8
  MAX_DIGIT_INDEX = DIGITS - 1

  MAX_DISPLAYABLE = 99_999_999
  HIGH_DIGIT_DIVIDER = 10_000_000
  
  SEGMENTS = 8
  MAX_SEGMENT_INDEX = SEGMENTS - 1
  
  BITS_TO_SHIFT = SEGMENTS + DIGITS
  MAX_BIT_INDEX = BITS_TO_SHIFT - 1

  SEGMENT_A = 1 << 7
  SEGMENT_B = 1 << 6
  SEGMENT_C = 1 << 5
  SEGMENT_D = 1 << 4
  SEGMENT_E = 1 << 3
  SEGMENT_F = 1 << 2
  SEGMENT_G = 1 << 1
  SEGMENT_DP = 1 << 0

  PAD_BITS_SPACE = 0
  PAD_BITS_ZERO = SEGMENT_A | SEGMENT_B | SEGMENT_C | SEGMENT_D | SEGMENT_E | SEGMENT_F
  PAD_BITS_UNDERLINE = SEGMENT_D

  DEFAULT_PAD_BITS = PAD_BITS_ZERO
                  
VAR

  long shiftStack[64] ' not optimized
  long fourDigitNumber, decimalDivider
  long cog
  long dataPin, clockPin, latchPin
  long padBits

  byte digitBits[DIGITS]

PUB Start(data, clock, latch, pad) 
'' This method starts a cog
'' "pad" is used to set the pad character.
'' If the character isn't listed.

  longmove(@dataPin, @data, 3)

  case pad
    " ":
      SetPadBits(PAD_BITS_SPACE)
    "0":
      SetPadBits(PAD_BITS_ZERO)
    "_":
      SetPadBits(PAD_BITS_UNDERLINE)
    other:
      SetPadBits(DEFAULT_PAD_BITS)
  
  result := cog := cognew(MaintainDigits, @shiftStack) + 1

PUB Set(valueToDisplay)

  fourDigitNumber := valueToDisplay
  
PUB SetPadBits(newPadBits)
'' Set the segments to light in unneeded digits.

  padBits := newPadBits
  
PRI MaintainDigits | previousFourDigitNumber
'' This method runs in its own cog. It will continuously
'' display the value of "fourDigitNumber" on the 7-segment display.

  dira[dataPin] := 1  ' direction settings need to be done within cog using the pins
  dira[clockPin] := 1
  dira[latchPin] := 1
 
  previousFourDigitNumber := $7FFFFFFF ' some impossible number
  
  repeat
    if fourDigitNumber <> previousFourDigitNumber ' check to see if it's a new number
      previousFourDigitNumber := fourDigitNumber
      BustDigits(fourDigitNumber)
           
    repeat result from 0 to MAX_DIGIT_INDEX  ' result is used as a temporary variable
      ShiftOutDigits(result, digitBits[result])
        
PUB BustDigits(numberToDisplay) | divisor, numberStartedFlag, singleDigitValue
'' Convert a number to bits representing the segments
'' to light in a 7-segment display.
 
  numberToDisplay <#= MAX_DISPLAYABLE
  divisor := HIGH_DIGIT_DIVIDER
  numberStartedFlag := 0
  
  repeat result from 0 to MAX_DIGIT_INDEX
    singleDigitValue := numberToDisplay / divisor
    if singleDigitValue or numberStartedFlag
      digitBits[result] := bitsInNumber[singleDigitValue]
      numberStartedFlag := 1
    else
      digitBits[result] := padBits
    numberToDisplay //= divisor
    divisor /= 10
    
PUB ShiftOutDigits(digit, activeBits) 
'' This methos moves the digit bit to the correct location and
'' combines the digit and segments bit together.
'' The combined bits are shifted out.
'' This method only displays a single digit at a time.
'' All 8 digits need to be cycled quickly to make them
'' all appear to be on.

      
  '' our numbering system 0 1 2 3 4 5 6 7
  '' numbering system of display 7 6 5 4 3 2 1 0
    
  digit := 1 << (7 - digit)  ' set the digit bit high
 
  digit &= $FF ' only keep 8 bits (this shouldn't really be needed)
  !activeBits ' invert segment bits, segments are cathodes
  activeBits &= $FF ' only keep 8 bits (this shouldn't really be needed) 
  activeBits <<= 8  ' move segment bits to high bits (last to be clocked out)
  activeBits |= digit  ' "Or" the bits controlling the digits together with the
                       ' bits controlling the segments.
                       ' The top 8 bits control the digits and the bottom 8 bits
                       ' control the segments.

  
  repeat BITS_TO_SHIFT
    outa[dataPin] := activeBits     
    outa[clockPin] := 1
    outa[clockPin] := 0
    activeBits >>= 1
  outa[latchPin] := 1
  outa[latchPin] := 0

DAT

bitsInNumber            byte SEGMENT_A | SEGMENT_B | SEGMENT_C | SEGMENT_D | SEGMENT_E | SEGMENT_F  '0
                        byte SEGMENT_B | SEGMENT_C  '1
                        byte SEGMENT_A | SEGMENT_B | SEGMENT_D | SEGMENT_E | SEGMENT_G '2
                        byte SEGMENT_A | SEGMENT_B | SEGMENT_C | SEGMENT_D | SEGMENT_G '3
                        byte SEGMENT_B | SEGMENT_C | SEGMENT_F | SEGMENT_G '4
                        byte SEGMENT_A | SEGMENT_C | SEGMENT_D | SEGMENT_F | SEGMENT_G '5
                        byte SEGMENT_A | SEGMENT_C | SEGMENT_D | SEGMENT_E | SEGMENT_F | SEGMENT_G '6
                        byte SEGMENT_A | SEGMENT_B | SEGMENT_C '7
                        byte SEGMENT_A | SEGMENT_B | SEGMENT_C | SEGMENT_D | SEGMENT_E | SEGMENT_F | SEGMENT_G '8
                        byte SEGMENT_A | SEGMENT_B | SEGMENT_C | SEGMENT_D | SEGMENT_F | SEGMENT_G '9
